#ifndef _Rep_info_h
#define _Rep_info_h 1

#include <sys/time.h>
#include "types.h"
#include "Time.h"
#include "Digest.h"
#include "Array.h"
#include "Reply.h"
#include "State_defs.h"

class Req_queue;

class Rep_info {
  //
  // Holds the last replies sent to each principal.
  //
public:
  Rep_info(char *mem, int sz, int nps);
  // Requires: "mem" points to an array of "size" bytes and is virtual
  // memory page aligned.
  // Effects: Creates a new object that stores data in "mem" for "nps"
  // principals.


  ~Rep_info();

  int size() const;
  // Effects: Returns the actual number of bytes (a multiple of the
  // virtual memory page size) that was consumed by this from the
  // start of the "mem" argument supplied to the constructor.

  Request_id req_id(int pid);
  // Requires: "pid" is a valid principal identifier.
  // Effects: Returns the timestamp in the last message sent to
  // principal "pid".

  Digest &digest(int pid);
  // Requires: "pid" is a valid principal identifier.
  // Effects: Returns a reference to the digest of the last reply
  // value sent to pid.

#ifdef SPECULATIVE
  Digest &history_digest(int pid);
  // Requires: "pid" is a valid principal identifier.
  // Effects: returns a reference to the history of the last reply
  // value sent to pid.

  Digest &history_digest();
  // Requires:  --
  // Effect:  Returns a reference to the current history digest

  void append_digest(Digest& h, Digest& d);
  // Requires: "dig" is a digest for a request
  // Effects: Sets the history digest to the result of history.combine(h, d);

  void set_history(Digest &h);
  // Effects:  sets the history to h


#endif

  Reply* reply(int pid);
  // Requires: "pid" is a valid principal identifier.
  // Effects: Returns a pointer to the last reply value sent to "pid"
  // or 0 if no such reply was sent.

  bool new_state(Req_queue *rset);
  // Effects: Updates this to reflect the new state and removes stale
  // requests from rset. If it removes the first request in "rset",
  // returns true; otherwise returns false.
  
#ifdef SPECULATIVE
  char *new_reply(int pid, Digest &history, Seqno seqno, int &size);
#else
  char *new_reply(int pid, int &size);
#endif
  // Requires: "pid" is a valid principal identifier.  
  // Effects: Returns a pointer to a buffer where the new reply value
  // for principal "pid" can be placed. The length of the buffer in bytes
  // is returned in "size". Sets the reply to tentative.

  void end_reply(int pid, Request_id rid, int size);
  // Requires: "pid" is a valid principal identifier.
  // Effects: Completes the construction of a new reply value: this is
  // informed that the reply value is size bytes long and computes its
  // digest.

  void commit_reply(int pid);
  // Requires: "pid" is a valid principal identifier.
  // Effects: Mark "pid"'s last reply committed.

  bool is_committed(int pid);
  // Requires: "pid" is a valid principal identifier.
  // Effects: Returns true iff the last reply sent to "pid" is
  // committed.

#ifdef SPECULATIVE
  void send_reply(int pid, View v, int id,  bool speculative = true);
#else
  void send_reply(int pid, View v, int id,  bool tentative=true);
#endif
  // Requires: "pid" is a valid principal identifier and end_reply was
  // called after the last call to new_reply for "pid"
  // Effects: Sends a reply message to "pid" for view "v" from replica
  // "id" with the latest reply value stored in the buffer returned by
  // new_reply. If tentative is omitted or true, it sends the reply as
  // tentative unless it was previously committed


private:
  int nps;
  char *mem;
#ifdef SPECULATIVE
  Digest *hist_dig;
  long mods;
#endif 
  Array<Reply*> reps; // Array of replies indexed by principal id.
  static const int Max_rep_size = 8192;

  struct Rinfo {
#ifdef SPECULATIVE
    bool speculative;  // True if last reply is speculative and not committed.
#else
    bool tentative;       // True if last reply is tentative and was not committed.
#endif
    Time lsent;           // Time at which reply was last sent.
  };
  Array<Rinfo> ireps;
};

inline int Rep_info::size() const {
  return (nps+1)*Max_rep_size;
}

inline void Rep_info::commit_reply(int pid) {
#ifdef SPECULATIVE
  ireps[pid].speculative = false;
#else
  ireps[pid].tentative = false;
#endif
  ireps[pid].lsent = zeroTime();
}

inline bool Rep_info::is_committed(int pid) {
#ifdef SPECULATIVE
  return !ireps[pid].speculative;
#else
  return !ireps[pid].tentative;
#endif
}

inline Request_id Rep_info::req_id(int pid) {
  return reps[pid]->request_id();
}

inline Digest& Rep_info::digest(int pid) {
  return reps[pid]->digest();
}

#ifdef SPECULATIVE
inline Digest& Rep_info::history_digest(int pid){
  return reps[pid]->history_digest();
}

inline Digest& Rep_info::history_digest(){
  return *hist_dig;
}


#endif

inline Reply* Rep_info::reply(int pid) {
  return reps[pid];
}


#endif // _Rep_info_h
